﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using Rhino;
using Rhino.Commands;
using Rhino.Display;
using Rhino.DocObjects;
using Rhino.Geometry;
using Rhino.Input.Custom;

namespace CustomRhinoObjectTest
{
    /// <summary>
    /// Just a sample enum
    /// </summary>
    public enum Colors { Black, Red, Green, Blue }

    /// <summary>
    /// Some basic class.
    /// It is the main object, the RhinoObject is a wrap/shell to be used from UI.
    /// 
    /// Entity  <---------->  RhinoObject 
    ///          \        /
    ///           UserData
    ///            
    /// Some properties can be changed from Rhino document (UI or viewport) or from code (using commands for example) and any change in RhinoObject/Entity must update the Entity/RhinoObject.
    /// </summary>
    public abstract class Entity : INotifyPropertyChanged
    {
        private object _value;
        private double _inflation;
        private Colors _color;
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Rhino object must know this property
        /// </summary>
        public abstract string Name { get; }
        /// <summary>
        /// Rhino object must know this property
        /// </summary>
        public Guid InstanceID { get; private set; }
        /// <summary>
        /// Rhino object must keep updated this property
        /// </summary>
        public object Value
        {
            get
            {
                return _value;
            }
            set
            {
                _value = value;
                OnPropertyChanged();
            }
        }
        /// <summary>
        /// This sample property can be changed from rhino or from code
        /// </summary>
        public double Inflation
        {
            get
            {
                return _inflation;
            }
            set
            {
                if (_inflation == value)
                    return;
                _inflation = value;
                OnPropertyChanged();
            }
        }
        /// <summary>
        /// This sample property can be changed from rhino or from code
        /// </summary>
        public Colors Color
        {
            get
            {
                return _color;
            }
            set
            {
                if (_color == value)
                    return;
                _color = value;
                OnPropertyChanged();
            }
        }

        public Entity()
        { 
            InstanceID = Guid.NewGuid();
            PropertyChanged += Entity_PropertyChanged; 
        }

        public  Entity Duplicate()
        {
            Entity e = (Entity)Activator.CreateInstance(GetType());
            e.InstanceID = InstanceID;
            e.Inflation = Inflation;
            e.Color = Color;
            if (Value != null && Value is Curve crv)
                e.Value = crv.DuplicateCurve();
            return e;
        }
        public override string ToString()
        {
            return $"{Name} [Inflation: {Inflation}, Color: {Color}, Value: {Value ?? "null"}]";
        }
        private void Entity_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            Rhino.RhinoApp.WriteLine($"OnPropertyChanged: {e.PropertyName}");
        }

        protected void OnPropertyChanged([System.Runtime.CompilerServices.CallerMemberName] string memberName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(memberName)); 
        }
    }

    /// <summary>
    /// An entity of type curve
    /// </summary>
    public class EntityCurve : Entity
    {
        public override string Name { get { return "EntityCurveName"; } }
        public new Curve Value
        {
            get
            {
                return (Curve)base.Value;
            }
            set
            {
                base.Value = value;
            }
        }

        public EntityCurve() : base()
        {

        }

        public RhinoObject ToRhinoObject()
        {
            return new RhinoEntityCurve(this);
        }
    }

    /// <summary>
    /// Rhino object must have UserData that can be changed from the document, how to achive that?
    /// </summary>
    public class EntityRhinoData : Rhino.DocObjects.Custom.UserDictionary
    {
        public override string Description => base.Description;
        public override bool ShouldWrite => base.ShouldWrite;

        public EntityRhinoData() : base() { }
        public EntityRhinoData(Entity entity) : base()
        {
            // Set entity data
            this.Dictionary.Set("InstanceID", entity.InstanceID);
            this.Dictionary.Set("PropInt", entity.Inflation);
        }

        protected override void OnDuplicate(Rhino.DocObjects.Custom.UserData source)
        {
            base.OnDuplicate(source);
        }
        protected override void OnTransform(Transform transform)
        {
            base.OnTransform(transform);
        }
        protected override void Dispose(bool disposing)
        {
            base.Dispose(disposing);
        }
        protected override bool Write(Rhino.FileIO.BinaryArchiveWriter archive)
        {
            return base.Write(archive);
        }
        protected override bool Read(Rhino.FileIO.BinaryArchiveReader archive)
        {
            return base.Read(archive);
        }
    }

    /// <summary>
    /// A shell of EntityCurve to be used in the document.
    /// </summary>
    public class RhinoEntityCurve : Rhino.DocObjects.Custom.CustomCurveObject
    {
        public EntityCurve Entity { get; private set; }

        public RhinoEntityCurve() : base() { }
        public RhinoEntityCurve(EntityCurve entity) : base()
        {
            Entity = entity;
            Name = entity.Name;
            Attributes.Name = entity.Name;
            Attributes.SetUserString("Inflation", entity.Inflation.ToString());
            Attributes.SetUserString("Color", entity.Color.ToString());
            Attributes.SetUserString("InstanceID", entity.InstanceID.ToString());

            // User data must be on the object or in its attributes in order to be changed from the document by the user?
            Attributes.UserData.Add(new EntityRhinoData(entity));
            var value = entity.Value;
            if (value != null)
            {
                SetCurve(value);
            }
            CommitChanges();
             
        }
        protected override void OnAddToDocument(RhinoDoc doc)
        {
            RhinoApp.WriteLine("OnAddToDocument RhinoEntityCurve with " + Entity.ToString());
            base.OnAddToDocument(doc);
        }
        protected override void OnDeleteFromDocument(RhinoDoc doc)
        {
            RhinoApp.WriteLine("OnDeleteFromDocument RhinoEntityCurve with " + Entity.ToString());
            base.OnDeleteFromDocument(doc);
        }
        protected override void OnSelectionChanged()
        {
            RhinoApp.WriteLine("OnSelectionChanged");
            base.OnSelectionChanged();
            RhinoPanel.OpenPanel();
            if (this.IsSelected(false) > 0)
            {
                RhinoPanel.Instance.SetEntity(Entity);
            }
            else
            {
                RhinoPanel.Instance.SetEntity(null);
            }
        }
        protected override void OnDraw(DrawEventArgs e)
        {
            base.OnDraw(e);
            if (this.IsSelected(false) > 0)
            {
                // Just some visual reference
                var bb = Geometry.GetBoundingBox(true);
                bb.Inflate(Entity.Inflation);
                Color color = Color.Empty;
                switch (Entity.Color)
                {
                    case Colors.Black:
                        color = Color.Black;
                        break;
                    case Colors.Red:
                        color = Color.Red;
                        break;
                    case Colors.Green:
                        color = Color.Green;
                        break;
                    case Colors.Blue:
                        color = Color.Blue;
                        break; 
                }
                e.Display.DrawBox(bb, color);

                // Visual reference of entity curve
                var pipe = Brep.CreatePipe(Entity.Value, 1, true, PipeCapMode.None, false, 0.01, 0.01);
                if(pipe != null)
                {
                    var mat = new DisplayMaterial(color);
                    foreach(var p in pipe)
                    {
                        e.Display.DrawBrepShaded(p, mat);
                    }
                }
               
            }
        }
        protected override void OnDuplicate(RhinoObject source)
        {
            RhinoApp.WriteLine("OnDuplicate");
            base.OnDuplicate(source); 
            if (source is RhinoEntityCurve rhe)
            {
                Entity = rhe.Entity.Duplicate() as EntityCurve; 
            } 
        }
        protected override void OnTransform(Transform transform)
        {
            RhinoApp.WriteLine("OnTransform");
            base.OnTransform(transform);
            Entity.Value = this.CurveGeometry;
        }
        protected override void OnSpaceMorph(SpaceMorph morph)
        {
            RhinoApp.WriteLine("OnSpaceMorph");
            base.OnSpaceMorph(morph);
            Entity.Value = this.CurveGeometry;
        }
        protected override void OnPicked(PickContext context, IEnumerable<ObjRef> pickedItems)
        {
            base.OnPicked(context, pickedItems);
        }
    
        // why override this trow an exception?
        // Rhino.Runtime.DocumentCollectedException: 'This object cannot be modified because it is controlled by a document.'
        //public override string ShortDescription(bool plural)
        //{
        //    return base.ShortDescription(plural);
        //}
    }
}